/**

	MC - Memory Controller
	
	File			:	Controller Relization

	Created by		:	Reznichenko Anton aka AlterVision
						mailto:me@av13.ru
						http://www.av13.ru

	Project started :	30 november 2008 - 17:00
	Last edited		:	31 november 2008 - 00:01

	More Info		:	http://www.av13.ru/arts/mc-heappie/

*/

#include <windows.h>
#include "mc.h"

#define null NULL				// Consuetudo est alter natura

// Heap Algorithm Constants
#define MC_ALGO_RND_NC	0		// Random Access Algorithm without minimization
#define MC_ALGO_RND_CL	1		// ... with minimization
#define	MC_ALGO_SEQ_CL	2		// Sequential Search Algorithm with minimization

// Heap Definition
typedef struct _heap heap_t, *heap_p;
struct _heap {
	size_t			freespace;		// Free Space
	byte *			map;			// Bitmap
	void *			data;			// Memory
	heap_p			next;			// Next Heap in Ring
};

// Heap Collection Type
typedef struct _heaps heaps_t, *heaps_p;
struct _heaps {
	size_t			min_s;			// Minimum Object Size
	size_t			max_s;			// Maximum Object Size
	heap_p			heap;			// Link to the Heap
	int				heap_size;		// Heap Size
	int				heap_cell;		// Heap Cell Size
	int				heap_algo;		// Heap Working Algo
};

// Pointers to Main Heaps
heap_p				heap4, heap8, heap128, heap1024, heap4096;

// Heaps Array
#define MC_HEAPS  5
heaps_t heaps[MC_HEAPS] = {
	{0,		4,			heap4,		32768,	4,		MC_ALGO_RND_NC},
	{4,		256,		heap8,		32768,	8,		MC_ALGO_RND_NC},
	{256,	4096,		heap128,	16384,	128,	MC_ALGO_RND_CL},
	{4096,	65536,		heap1024,	2048,	1024,	MC_ALGO_SEQ_CL},
	{65536,	524288,		heap4096,	1024,	4096,	MC_ALGO_SEQ_CL},
};

//
// mc_ceil_div - common ceil division functions
int __inline mc_ceil_div (int a, int b) {
	int c;
	c = a / b;
	if (a % b) ++c;
	return c;
}
int __inline mc_ceil_blocks (size_t a, int b) {
	int c;
	c = (int) (a / b);
	if (a % b) ++c;
	return c;
}
// end :: mc_ceil_div
//

//
// mc_new_heap - creates new heap in memory
heap_p mc_new_heap (heaps_p root, heap_p h_prev = null) {

	int			i;
	size_t		h_size;
	heap_p		h_new;
	byte *		temp_ptr;

	// Allocate Memory for the Full Heap
	h_size = sizeof (heap_t) + (root->heap_size * (root->heap_cell + 1));
	h_new = (heap_p) HeapAlloc (GetProcessHeap (), 0, h_size);
	if (!h_new) return null;

	// Setting Link to Bit Map
	temp_ptr = (byte*) h_new;
	temp_ptr += sizeof (heap_t);
	h_new->map = temp_ptr;

	// Setting Link to Data Space
	temp_ptr += root->heap_size;
	h_new->data = (void*) temp_ptr;

	// Configure Heap
	for (i = 0; i < root->heap_size; ++i) {
		h_new->map[i] = 0;
	}

	h_new->freespace = root->heap_size * root->heap_cell;
	
	// References
	if (h_prev) {
		// Setting New Heap Right Afret the Sent
		h_new->next = (h_prev->next) ? h_prev->next : h_prev;
		h_prev->next = h_new;
	} else {
		// Creating Standalone Heap
		h_new->next = null;
	}

	return h_new;

}
// end :: mc_new_heap
//

//
// mc_kill_heap
void mc_kill_heap (heaps_p root, heap_p heap_to_die) {

    heap_p h_prev;

	// If Heap Is Not The Last One
	if (heap_to_die->next) {

		// Searching For Previous Heap in Ring
		h_prev = heap_to_die->next;
		while (h_prev->next != heap_to_die) {
			h_prev = h_prev->next;
		}

		// Reset References
		if (h_prev == heap_to_die->next) {
			// The Only Heap
			h_prev->next = null;
		} else {
			// Change Heap's References
			h_prev->next = heap_to_die->next;
		}

		// Change Reference of Main Heap
		if (heap_to_die == root->heap) {
			root->heap = heap_to_die->next;
		}

		// Kill the Heappie ...
		HeapFree  (GetProcessHeap (), 0, heap_to_die);

	}
	// Heappie not dead!! =)

}
// end :: mc_kill_heap
//

//
// mc_heap_alloc - searches along the heap to find free space
void * mc_heap_alloc (heaps_p root, heap_p search_heap, size_t size) {

	void *	out_p;				// Outer Variable Pointer
	byte	*seek_p;	// Seeker Variable Pointer
	int		size_b, size_tmp;	// Size in Blocks and Blocks in Size =)
	int		i, j, k, t;			// Iterators
	int		st, en;				// Start and End Pointers
	int		maxe;				// Start / End Pointers Changed
	int		place_found;		// Found place for new variable
	heap_p	reserve;			// Reserved Heap

	if (search_heap->freespace < size) {
		// Search Elsewhere if Heap is Too Small
		if (search_heap->next && search_heap->next != root->heap) {
			// Go to the next heap
			out_p = mc_heap_alloc (root, search_heap->next, size);
		} else {
			// Create new heap and start there
			reserve = mc_new_heap (root, search_heap);
			out_p = mc_heap_alloc (root, reserve, size);
		}
	} else {
		// Prepare
		place_found = -1;
		size_b = mc_ceil_blocks (size, root->heap_cell);
		t = mc_ceil_div (root->heap_size, size_b);
		i = (root->heap_algo != MC_ALGO_SEQ_CL) ? (rand() % root->heap_size) : 0;

		// Start Search
		for (k = 0 ; k < t; k++) {
			// If Free Space Cell Found
			if (search_heap->map[i] == 0) {
				
				// For single blocks
				if (size_b == 1) {
					place_found = i;
					break;
				}

				// For blocks with large size
				size_tmp = size_b - 1;

				// Moving Backwards
				st = en = i;
				while (1) {
					if (st < 1) break;
					if (search_heap->map[st - 1] == 0) {
						--st;
						--size_tmp;
						if (size_tmp < 1) break;
					} else break;
				} 

				// Checking Exit Flag
				if (size_tmp < 1) {
					place_found = st;
					break;
				}

				maxe = root->heap_size - 2;
				// Moving Forward
				while (1) {
					if (en > maxe) break;
					if (search_heap->map[en + 1] == 0) {
						++en;
						--size_tmp;
						if (size_tmp < 1) break;
					} else break;
				} 

				// Checking Exit Flag
				if (size_tmp < 1) {
					place_found = st;
					break;
				}
				
			}
			// Coming to the Next Cell
			i = (i + size_b) % root->heap_size;
			if (place_found != -1) break;
		}


		// Allocate Space
		if (place_found != -1) {
			// In This Heap - mark place as used
			j = place_found;
			for (i = size_b; i > 0; --i) {
				search_heap->map[j] = min (2, i);
				j++;
			}
			// Find Pointer to Variable Location
			seek_p = (byte*) search_heap->data;
			seek_p += (place_found * root->heap_cell);
			out_p = (void*) seek_p;
			// Substract Heap Free Space
			search_heap->freespace -= size_b * root->heap_cell;
			// Change Heap Root to Current Heap
			root->heap = search_heap;
		} else {
			// Search Elsewhere if Heap is Too Small
			if (search_heap->next && search_heap->next != root->heap) {
				// Go to the next heap
				out_p = mc_heap_alloc (root, search_heap->next, size);
			} else {
				// Create new heap and start there
				reserve = mc_new_heap (root, search_heap);
				out_p = mc_heap_alloc (root, reserve, size);
			}
		}

	}

	return out_p;

}
// end :: mc_heap_alloc
//

//
// mc_heap_delete - deleted pointer from heap
void mc_heap_delete (heaps_p root, heap_p delete_heap, int offset) {

	int		i;
	byte *	bitmap_p;

	// Walk Through the Bitmap and Clear All
	bitmap_p = delete_heap->map + offset;
	do {
		i = *bitmap_p;	
		*bitmap_p = 0;
		++bitmap_p;
		--i;
	} while (i);

	delete_heap->freespace += root->heap_cell * ( (bitmap_p - delete_heap->map) - offset );

	// If MC is configured to kill empty heappies ...
	if ( root->heap_algo == MC_ALGO_RND_CL || root->heap_algo == MC_ALGO_SEQ_CL ) {
		if (delete_heap->freespace == root->heap_cell * root->heap_size) {
			mc_kill_heap (root, delete_heap);
		}
	}

}
// end :: mc_heap_delete
//

//
// mc_alloc - checks heap and initializes search
void * mc_alloc ( size_t size ) {

	int i;						// Heap Iterator

	if (size == 0) return null;	// No jokes ...

	// Walk Through Heaps Array and Find Suitable Root
	for (i = 0; i < MC_HEAPS; ++i) {
		// If suitable array found - allocate memory in it
		if (size > heaps[i].min_s && size <= heaps[i].max_s) {
			// Creating Heap Root if Necessary
			if (heaps[i].heap == null) {
				heaps[i].heap = mc_new_heap(&heaps[i]);
				if (heaps[i].heap == null) return null;
			}
			return mc_heap_alloc (&heaps[i], heaps[i].heap, size);
		}
	}

	// If the variable is too big, use Windows HeapAlloc function
	return HeapAlloc (GetProcessHeap (), 0, size);

}
// end :: mc_alloc

//
// mc_delete - kills the memory ... hard, but efficient
void mc_delete ( void * mc_block ) {

	byte	*tp1, *tp2, *tp3;	// Temporary Byte Pointers
	heap_p	hp;					// Temporary Heappie Pointer
	int		offset, i;			// Iterator Variables

	tp1 = (byte*) mc_block;		// Byte position of memory block
	offset = -1;

	// Walk Through Heaps Array and Find Suitable Root
	for (i = 0; i < MC_HEAPS; ++i) {
		// Look throuth the heap
		hp = heaps[i].heap; // from the beginning of the ring
		if (hp == null) continue;
		do {
			// Byte Bounds of Heappie Data Field
			tp2 = (byte*) hp->data;
			tp3 = tp2 + heaps[i].heap_size * heaps[i].heap_cell;
	        // Counting Block Offset
			if (tp1 >= tp2 && tp1 <= tp3) {
				offset = (int) (tp1 - tp2);
				offset /= heaps[i].heap_cell;
				break;
			}
			// Go to Next Heappie in Ring if Possible
			if (hp->next) {
				hp = hp->next;	
			} else {
				break;
			}
		} while (hp != heaps[i].heap);	// ... until it's begining again
		// if offset found, kill variable in the heap
		if (offset != -1) {
			mc_heap_delete (&heaps[i], hp, offset);
			mc_block = null;
			return;
		}
	}

	// If offset was not found, it was generated with Windows HeapAlloc,
	// so we delete it with Win HeapFree ...
	HeapFree (GetProcessHeap (), 0, mc_block);

}
// end :: mc_delete
//
